#!/usr/bin/env python
import logging
from pathlib import Path

from edc_test_utils import DefaultTestSettings, func_main

app_name = "edc_visit_tracking"


base_dir = Path(__file__).absolute().parent

project_settings = DefaultTestSettings(
    calling_file=__file__,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    SITE_ID=1,
    ETC_DIR=str(base_dir / app_name / "tests" / "etc"),
    SUBJECT_VISIT_MODEL="edc_visit_tracking.subjectvisit",
    EDC_LTFU_MODEL_NAME="edc_ltfu.ltfu",
    EDC_SITES_REGISTER_DEFAULT=True,
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.messages",
        "django.contrib.sessions",
        "django.contrib.sites",
        "django.contrib.staticfiles",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "multisite",
        "edc_action_item.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_crf.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_facility.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_list_data.apps.AppConfig",
        "edc_ltfu.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_protocol.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_sites.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
    ],
    add_dashboard_middleware=True,
).settings


def main():
    func_main(project_settings, *[f"{app_name}.tests"])


if __name__ == "__main__":
    logging.basicConfig()
    main()



from datetime import datetime

from django.apps import apps as django_apps
from edc_appointment.creators import UnscheduledAppointmentCreator
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules


class Helper:
    def __init__(
        self,
        subject_identifier=None,
        subject_consent_model_cls=None,
        onschedule_model_name=None,
    ):
        self.subject_identifier = subject_identifier
        self.subject_consent_model_cls = subject_consent_model_cls or django_apps.get_model(
            "edc_visit_tracking.subjectconsent"
        )
        self.onschedule_model_name = (
            onschedule_model_name or "edc_visit_tracking.onscheduleone"
        )

    def consent_and_put_on_schedule(
        self, subject_identifier: str | None = None, consent_datetime: datetime | None = None
    ):
        subject_identifier = subject_identifier or self.subject_identifier
        consent_datetime = consent_datetime or get_utcnow()
        subject_consent = self.subject_consent_model_cls.objects.create(
            subject_identifier=subject_identifier, consent_datetime=consent_datetime
        )
        _, schedule = site_visit_schedules.get_by_onschedule_model(self.onschedule_model_name)
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        return subject_consent

    @staticmethod
    def create_unscheduled(appointment):
        return UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
            facility=appointment.facility,
        ).appointment



from django.contrib import admin
from django.test import TestCase
from django.test.client import RequestFactory
from django_audit_fields.admin import ModelAdminAuditFieldsMixin, audit_fields
from edc_appointment.models import Appointment
from edc_facility.import_holidays import import_holidays
from edc_visit_schedule.fieldsets import visit_schedule_fields
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.admin_site import edc_visit_tracking_admin
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.modeladmin_mixins import CrfModelAdminMixin
from edc_visit_tracking.models import SubjectVisit

from ..helper import Helper
from ..models import CrfOne
from ..visit_schedule import visit_schedule1, visit_schedule2


@admin.register(CrfOne, site=edc_visit_tracking_admin)
class CrfOneModelAdmin(CrfModelAdminMixin, ModelAdminAuditFieldsMixin, admin.ModelAdmin):
    def get_field_queryset(self, db, db_field, request):
        return CrfOne.objects.all()


class TestModelAdmin(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)

    def test_related_visit_model_attr(self):
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)
        self.assertTrue("subject_visit", modeladmin.related_visit_model_attr)

    def test_adds_to_list_display(self):
        factory = RequestFactory()
        request = factory.get("/")
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)
        self.assertIn("subject_identifier", modeladmin.get_list_display(request))
        self.assertIn("report_datetime", modeladmin.get_list_display(request))

    def test_extends_list_filter(self):
        factory = RequestFactory()
        request = factory.get("/")
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)
        self.assertIn(
            f"{modeladmin.related_visit_model_attr}__report_datetime",
            modeladmin.get_list_filter(request),
        )
        self.assertIn(
            f"{modeladmin.related_visit_model_attr}__reason",
            modeladmin.get_list_filter(request),
        )
        self.assertIn(
            f"{modeladmin.related_visit_model_attr}__visit_code",
            modeladmin.get_list_filter(request),
        )
        self.assertIn(
            f"{modeladmin.related_visit_model_attr}__visit_code_sequence",
            modeladmin.get_list_filter(request),
        )

    def test_extends_search_fields(self):
        factory = RequestFactory()
        request = factory.get("/")
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)
        self.assertIn(
            f"{modeladmin.related_visit_model_attr}__appointment__subject_identifier",
            modeladmin.get_search_fields(request),
        )

    def test_extends_fk_none(self):
        factory = RequestFactory()
        request = factory.get(
            "/?next=my_url_name,arg1,arg2&arg1=value1&arg2=value2&arg3=value3&arg4=value4"
        )
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)

        class Fld:
            name = modeladmin.related_visit_model_attr

            def formfield(self, **kwargs):
                return kwargs

        db_field = Fld()

        kwargs = modeladmin.formfield_for_foreignkey(db_field, request)
        self.assertEqual(kwargs["queryset"].count(), 0)

    def test_extends_fk_not_none(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint_datetime")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)

        factory = RequestFactory()
        request = factory.get(
            f"/?next=my_url_name,arg1,arg2&arg1=value1&arg2=value2&"
            f"subject_visit={str(subject_visit.id)}&arg4=value4"
        )
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)

        CrfOne.objects.create(subject_visit=subject_visit)

        class Fld:
            name = modeladmin.related_visit_model_attr

            def formfield(self, **kwargs):
                return kwargs

        db_field = Fld()
        kwargs = modeladmin.formfield_for_foreignkey(db_field, request)
        self.assertGreater(kwargs["queryset"].count(), 0)

    def test_visit_extends_fk_not_none(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint_datetime")[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)

        factory = RequestFactory()
        request = factory.get(
            f"/?next=my_url_name,arg1,arg2&arg1=value1&arg2=value2"
            f"&appointment={str(appointment.id)}"
        )
        modeladmin = edc_visit_tracking_admin._registry.get(SubjectVisit)

        class Fld:
            name = "appointment"

            related_model = Appointment

            def formfield(self, **kwargs):
                return kwargs

        db_field = Fld()
        kwargs = modeladmin.formfield_for_foreignkey(db_field, request)
        self.assertGreater(kwargs["queryset"].count(), 0)

    def test_crf_readonly(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint_datetime")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)

        crf_one = CrfOne.objects.create(subject_visit=subject_visit)

        factory = RequestFactory()
        request = factory.get(
            f"/?next=my_url_name,arg1,arg2&arg1=value1&arg2=value2"
            f"&subject_visit={str(subject_visit.id)}"
        )
        modeladmin = edc_visit_tracking_admin._registry.get(CrfOne)
        for field in audit_fields:
            self.assertIn(field, modeladmin.get_readonly_fields(request, obj=crf_one))

    def test_visit_readonly(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint_datetime")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)

        factory = RequestFactory()
        request = factory.get(
            f"/?next=my_url_name,arg1,arg2&arg1=value1&arg2=value2"
            f"&appointment={str(appointment.id)}"
        )
        modeladmin = edc_visit_tracking_admin._registry.get(SubjectVisit)
        for field in audit_fields:
            self.assertIn(field, modeladmin.get_readonly_fields(request, obj=subject_visit))
        for field in visit_schedule_fields:
            self.assertIn(field, modeladmin.get_readonly_fields(request, obj=subject_visit))



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from edc_appointment.constants import MISSED_APPT
from edc_appointment.models import Appointment
from edc_constants.constants import (
    ALIVE,
    COMPLETE,
    NO,
    NOT_APPLICABLE,
    ON_STUDY,
    PARTICIPANT,
    YES,
)
from edc_facility.import_holidays import import_holidays
from edc_metadata.constants import MISSED
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.form_validators import VisitFormValidator
from edc_visit_tracking.modelform_mixins import (
    VisitTrackingCrfModelFormMixin,
    VisitTrackingModelFormMixin,
)
from edc_visit_tracking.models import SubjectVisit

from ...exceptions import RelatedVisitReasonError
from ..helper import Helper
from ..models import BadCrfNoRelatedVisit, CrfOne
from ..visit_schedule import visit_schedule1, visit_schedule2

utc_tz = ZoneInfo("UTC")


class SubjectVisitForm(VisitTrackingModelFormMixin, forms.ModelForm):
    form_validator_cls = VisitFormValidator

    class Meta:
        model = SubjectVisit
        fields = "__all__"


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestForm(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)

    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2022, 6, 11, 8, 00, tzinfo=utc_tz),
    )
    @time_machine.travel(datetime(2019, 7, 12, 8, 00, tzinfo=utc_tz))
    def test_visit_tracking_form_ok(self):
        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 6, 12, 8, 00, tzinfo=utc_tz)
        )

        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]

        data = dict(
            appointment=appointment,
            subject_identifier=appointment.subject_identifier,
            report_datetime=appointment.appt_datetime,
            info_source=PARTICIPANT,
            study_status=ON_STUDY,
            survival_status=ALIVE,
            reason=SCHEDULED,
            reason_unscheduled=NOT_APPLICABLE,
            document_status=COMPLETE,
            require_crfs=YES,
        )
        form = SubjectVisitForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_visit_tracking_form__appt_not_missed_but_visit_missed(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]

        data = dict(
            appointment=appointment,
            subject_identifier=appointment.subject_identifier,
            report_datetime=appointment.appt_datetime,
            info_source=PARTICIPANT,
            study_status=ON_STUDY,
            survival_status=ALIVE,
            reason=MISSED,
            reason_unscheduled=NOT_APPLICABLE,
            document_status=COMPLETE,
            require_crfs=YES,
        )
        form = SubjectVisitForm(data=data)
        form.is_valid()
        self.assertIn("reason", form._errors)

    def test_visit_tracking_form__appt_missed_but_visit_not_missed(self):
        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 5, 11, 00, tzinfo=utc_tz)
        )
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]

        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()

        data = dict(
            appointment=appointment,
            subject_identifier=appointment.subject_identifier,
            report_datetime=appointment.appt_datetime,
            info_source=PARTICIPANT,
            study_status=ON_STUDY,
            survival_status=ALIVE,
            reason=SCHEDULED,
            reason_unscheduled=NOT_APPLICABLE,
            document_status=COMPLETE,
            require_crfs=YES,
        )
        form = SubjectVisitForm(data=data)
        form.is_valid()
        self.assertIn("reason", form._errors)

    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2022, 6, 11, 8, 00, tzinfo=utc_tz),
    )
    @time_machine.travel(datetime(2019, 7, 12, 8, 00, tzinfo=utc_tz))
    def test_visit_tracking_form__missed_ok(self):
        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 6, 12, 8, 00, tzinfo=utc_tz)
        )
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]

        SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]

        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]

        # note if appt_timing = MISSED_APPT, will auto create subject visit
        # see VisitModelManager
        subject_visit = SubjectVisit.objects.get(appointment=appointment)

        data = dict(
            appointment=appointment,
            subject_identifier=appointment.subject_identifier,
            report_datetime=appointment.appt_datetime,
            info_source=PARTICIPANT,
            study_status=ON_STUDY,
            survival_status=ALIVE,
            reason=MISSED,
            reason_unscheduled=NOT_APPLICABLE,
            document_status=COMPLETE,
            require_crfs=YES,
        )
        form = SubjectVisitForm(data=data, instance=subject_visit)
        form.is_valid()
        form.save(commit=True)
        obj = SubjectVisit.objects.get(appointment=appointment)
        self.assertEqual(NO, obj.require_crfs)

    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2022, 6, 11, 8, 00, tzinfo=utc_tz),
    )
    @time_machine.travel(datetime(2019, 7, 12, 8, 00, tzinfo=utc_tz))
    def test_visit_tracking_model__appt_missed_but_visit_not_missed(self):
        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 6, 12, 00, tzinfo=utc_tz)
        )
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]

        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()

        self.assertRaises(
            RelatedVisitReasonError,
            SubjectVisit.objects.create,
            appointment=appointment,
            reason=SCHEDULED,
        )

    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2022, 6, 11, 8, 00, tzinfo=utc_tz),
    )
    @time_machine.travel(datetime(2019, 7, 12, 8, 00, tzinfo=utc_tz))
    def test_crf_with_visit_tracking_form_ok(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 6, 12, 00, tzinfo=utc_tz)
        )
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": get_utcnow(),
                "subject_visit": subject_visit.pk,
            }
        )

        self.assertTrue(form.is_valid())
        form.save(commit=True)

    def test_crf_with_visit_tracking_form_missing_subject_visit(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm({"f1": "1", "f2": "2", "f3": "3", "report_datetime": get_utcnow()})
        form.is_valid()
        self.assertIn("subject_visit", form._errors)

    def test_crf_with_visit_tracking_form_missing_subject_visit_fk_raises(self):
        class BadCrfNoRelatedVisitorm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = BadCrfNoRelatedVisit
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        form = BadCrfNoRelatedVisitorm(
            {"f1": "1", "f2": "2", "f3": "3", "report_datetime": get_utcnow()}
        )
        self.assertRaises(ImproperlyConfigured, form.is_valid)

    def test_crf_with_visit_tracking_form_no_report_datetime(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm({"f1": "1", "f2": "2", "f3": "3", "subject_visit": subject_visit.pk})
        self.assertFalse(form.is_valid())
        self.assertIn("report_datetime", form._errors)

    def test_crf_with_visit_tracking_form_report_datetime_validated_against_related_visit(
        self,
    ):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        for report_datetime in [
            get_utcnow() - relativedelta(months=1),
            get_utcnow() + relativedelta(months=1),
        ]:
            form = CrfForm(
                {
                    "f1": "1",
                    "f2": "2",
                    "f3": "3",
                    "report_datetime": report_datetime,
                    "subject_visit": subject_visit.pk,
                }
            )
            self.assertFalse(form.is_valid())
            self.assertIn("report_datetime", form._errors)

    @override_settings(TIME_ZONE="Africa/Dar_es_Salaam")
    def test_crf_with_visit_tracking_form_report_datetime_zone(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=get_utcnow(),
        )
        dte = get_utcnow().astimezone(ZoneInfo("Africa/Dar_es_Salaam"))
        for report_datetime in [
            dte - relativedelta(months=1),
            dte + relativedelta(months=1),
        ]:
            form = CrfForm(
                {
                    "f1": "1",
                    "f2": "2",
                    "f3": "3",
                    "report_datetime": report_datetime,
                    "subject_visit": subject_visit.pk,
                }
            )
            self.assertFalse(form.is_valid())
            self.assertIn("report_datetime", form._errors)

    @override_settings(
        EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz),
        EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2022, 6, 11, 8, 00, tzinfo=utc_tz),
        TIME_ZONE="Africa/Dar_es_Salaam",
    )
    @time_machine.travel(datetime(2019, 7, 12, 8, 00, tzinfo=utc_tz))
    def test_crf_with_visit_tracking_form_report_datetime_zone2(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule(
            consent_datetime=datetime(2019, 6, 12, 8, 00, tzinfo=utc_tz)
        )
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": subject_visit.report_datetime,
                "subject_visit": subject_visit.pk,
            }
        )
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save(commit=True)

        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": subject_visit.report_datetime,
                "subject_visit": subject_visit.pk,
            }
        )
        form.is_valid()
        self.assertIn("subject_visit", form._errors)



from .signals import (
    subject_visit_missed_on_post_delete,
    visit_tracking_check_in_progress_on_post_save,
)
from .subject_visit import SubjectVisit
from .subject_visit_missed import SubjectVisitMissed
from .subject_visit_missed_reasons import SubjectVisitMissedReasons
from .visit_reasons import VisitReasons



from __future__ import annotations

from edc_appointment.tests.test_case_mixins import AppointmentTestCaseMixin
from edc_consent.tests.test_case_mixins import ConsentTestCaseMixin
from edc_visit_schedule.constants import DAY1

from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.utils import get_related_visit_model_cls


class VisitTrackingTestCaseMixin(AppointmentTestCaseMixin, ConsentTestCaseMixin):
    def get_subject_visit(
        self,
        visit_code=None,
        visit_code_sequence=None,
        reason=None,
        appt_datetime=None,
    ):
        subject_visit_model_cls = get_related_visit_model_cls()
        reason = reason or SCHEDULED
        subject_consent = self.get_subject_consent()
        options = dict(
            subject_identifier=subject_consent.subject_identifier,
            visit_code=visit_code or DAY1,
            visit_code_sequence=(
                visit_code_sequence if visit_code_sequence is not None else 0
            ),
            reason=reason,
        )
        if appt_datetime:
            options.update(appt_datetime=appt_datetime)
        appointment = self.get_appointment(**options)
        subject_visit = subject_visit_model_cls(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=appointment.appt_datetime,
        )
        subject_visit.save()
        subject_visit.refresh_from_db()
        return subject_visit



from .caretaker_fields_mixin import CaretakerFieldsMixin
from .previous_visit_model_mixin import PreviousVisitError, PreviousVisitModelMixin
from .visit_model_fields_mixin import VisitModelFieldsMixin
from .visit_model_mixin import VisitModelMixin



# Generated by Django 4.2.5 on 2023-09-12 15:16

import uuid

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import edc_metadata.metadata_helper.metadata_helper_mixin
import edc_metadata.model_mixins.creates.creates_metadata_model_mixin
import edc_model.models.fields.other_charfield
import edc_model.validators.date
import edc_model_fields.fields.other_charfield
import edc_protocol.validators
import edc_sites.models
import edc_utils.date
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_visit_tracking.managers


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_appointment", "0042_alter_appointment_device_created_and_more"),
        ("sites", "0002_alter_domain_unique"),
        ("edc_visit_tracking", "0003_auto_20220913_2139"),
    ]

    operations = [
        migrations.CreateModel(
            name="SubjectVisit",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                ("visit_code", models.CharField(editable=False, max_length=25, null=True)),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "visit_schedule_name",
                    models.CharField(
                        editable=False,
                        help_text='the name of the visit schedule used to find the "schedule"',
                        max_length=25,
                    ),
                ),
                ("schedule_name", models.CharField(editable=False, max_length=25)),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=edc_utils.date.get_utcnow,
                        help_text="Date and time of this report",
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.validators.date.datetime_not_future,
                        ],
                        verbose_name="Report date and time",
                    ),
                ),
                (
                    "reason_unscheduled",
                    models.CharField(
                        choices=[
                            ("patient_unwell_outpatient", "Patient unwell (outpatient)"),
                            ("patient_hospitalised", "Patient hospitalised"),
                            ("OTHER", "Other"),
                            ("N/A", "Not applicable"),
                        ],
                        default="N/A",
                        max_length=25,
                        verbose_name="If 'unscheduled', provide the reason for the unscheduled visit",
                    ),
                ),
                (
                    "reason_unscheduled_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=25,
                        null=True,
                        verbose_name='If the reason for the unscheduled visit is "other", specify',
                    ),
                ),
                (
                    "reason_missed_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=25,
                        null=True,
                        verbose_name='If the reason for the missed visit is "other", specify',
                    ),
                ),
                (
                    "study_status",
                    models.CharField(
                        max_length=50,
                        null=True,
                        verbose_name="What is the participant's current study status",
                    ),
                ),
                (
                    "require_crfs",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No")],
                        default="Yes",
                        max_length=10,
                        verbose_name="Are scheduled data being submitted with this visit?",
                    ),
                ),
                (
                    "info_source_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name='If "Other" source of information, specify',
                    ),
                ),
                (
                    "survival_status",
                    models.CharField(
                        choices=[
                            ("alive", "Alive"),
                            ("dead", "Deceased"),
                            ("unknown", "Unknown"),
                            ("N/A", "Not applicable (if missed)"),
                        ],
                        default="alive",
                        help_text="If YES, submit Death report",
                        max_length=10,
                        null=True,
                        verbose_name="Participant's survival status",
                    ),
                ),
                (
                    "last_alive_date",
                    models.DateField(
                        blank=True,
                        help_text="Will be validated on the Death report",
                        null=True,
                        validators=[
                            edc_protocol.validators.date_not_before_study_start,
                            edc_model.validators.date.date_not_future,
                        ],
                        verbose_name="Date participant last known alive",
                    ),
                ),
                (
                    "comments",
                    models.TextField(
                        blank=True,
                        max_length=250,
                        null=True,
                        verbose_name="Comment if any additional pertinent information about the participant",
                    ),
                ),
                (
                    "document_status",
                    models.CharField(
                        choices=[
                            ("INCOMPLETE", "Incomplete (some data pending)"),
                            ("COMPLETE", "Complete"),
                        ],
                        default="COMPLETE",
                        help_text="If some data is still pending, flag as incomplete",
                        max_length=25,
                        verbose_name="Document status",
                    ),
                ),
                (
                    "document_status_comments",
                    models.TextField(
                        blank=True,
                        help_text="for example, why some data is still pending",
                        null=True,
                        verbose_name="Any comments related to status of this document",
                    ),
                ),
                ("consent_model", models.CharField(editable=False, max_length=50, null=True)),
                (
                    "consent_version",
                    models.CharField(editable=False, max_length=10, null=True),
                ),
                (
                    "reason",
                    models.CharField(
                        choices=[
                            ("scheduled", "Scheduled visit/contact"),
                            ("unscheduled", "Unscheduled visit/contact"),
                            ("missed", "Missed visit"),
                            (
                                "lost",
                                "Lost to follow-up (use only when taking subject off study)",
                            ),
                            ("deferred", "Deferred"),
                            ("completed protocol", "Completed protocol"),
                        ],
                        max_length=25,
                    ),
                ),
                (
                    "reason_missed",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("timepoint", "Missed timepoint"),
                            ("OTHER", "Other"),
                            ("N/A", "Not applicable"),
                        ],
                        max_length=35,
                        null=True,
                        verbose_name="If 'missed', provide the reason for the missed visit",
                    ),
                ),
                (
                    "info_source",
                    models.CharField(
                        choices=[
                            ("participant", "1. Clinic visit with participant"),
                            ("other_contact", "2. Other contact with participant"),
                            (
                                "other_doctor",
                                "3. Contact with external health care provider/medical doctor",
                            ),
                            (
                                "family",
                                "4. Contact with family or designated person who can provide information",
                            ),
                            ("chart", "5. Hospital chart or other medical record"),
                            ("OTHER", "9. Other"),
                        ],
                        max_length=25,
                        verbose_name="What is the main source of this information?",
                    ),
                ),
                (
                    "appointment",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="default_subjectvisit",
                        to="edc_appointment.appointment",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "ordering": (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "report_datetime",
                ),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
                "default_manager_name": "objects",
            },
            bases=(
                edc_metadata.metadata_helper.metadata_helper_mixin.MetadataHelperMixin,
                edc_metadata.model_mixins.creates.creates_metadata_model_mixin.RelatedVisitProtocol,
                models.Model,
            ),
            managers=[
                ("objects", edc_visit_tracking.managers.VisitModelManager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AddField(
            model_name="subjectvisitmissedreasons",
            name="extra_value",
            field=models.CharField(max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="visitreasons",
            name="extra_value",
            field=models.CharField(max_length=250, null=True),
        ),
        migrations.CreateModel(
            name="HistoricalSubjectVisitMissed",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=edc_utils.date.get_utcnow,
                        help_text="If reporting today, use today's date/time, otherwise use the date/time this information was reported.",
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.validators.date.datetime_not_future,
                        ],
                        verbose_name="Report Date",
                    ),
                ),
                (
                    "survival_status",
                    models.CharField(
                        choices=[
                            ("alive", "Alive"),
                            ("dead", "Deceased"),
                            ("unknown", "Unknown"),
                        ],
                        help_text="If deceased, complete the death report",
                        max_length=25,
                        verbose_name="Survival status",
                    ),
                ),
                (
                    "contact_attempted",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No")],
                        help_text="Not including pre-appointment reminders",
                        max_length=25,
                        verbose_name="Were any attempts made to contact the participant since the expected appointment date?",
                    ),
                ),
                (
                    "contact_attempts_count",
                    models.IntegerField(
                        blank=True,
                        help_text="Not including pre-appointment reminders. Multiple attempts on the same day count as a single attempt.",
                        null=True,
                        validators=[django.core.validators.MinValueValidator(1)],
                        verbose_name="Number of attempts made to contact participantsince the expected appointment date",
                    ),
                ),
                (
                    "contact_attempts_explained",
                    models.TextField(
                        blank=True,
                        null=True,
                        verbose_name="If contact not made and less than 3 attempts, please explain",
                    ),
                ),
                (
                    "contact_last_date",
                    models.DateField(
                        blank=True,
                        null=True,
                        validators=[
                            edc_model.validators.date.date_not_future,
                            edc_protocol.validators.date_not_before_study_start,
                        ],
                        verbose_name="Date of last telephone contact/attempt",
                    ),
                ),
                (
                    "contact_made",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No"), ("N/A", "Not applicable")],
                        default="N/A",
                        max_length=25,
                        verbose_name="Was contact finally made with the participant?",
                    ),
                ),
                (
                    "missed_reasons_other",
                    edc_model.models.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name="If other, please specify ...",
                    ),
                ),
                (
                    "ltfu",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No"), ("N/A", "Not applicable")],
                        default="No",
                        help_text="If 'Yes', complete the Loss to Follow up form",
                        max_length=15,
                        verbose_name="Has the participant met the protocol criteria for lost to follow up?",
                    ),
                ),
                (
                    "comment",
                    models.TextField(
                        blank=True,
                        null=True,
                        verbose_name="Please provide further details, if any",
                    ),
                ),
                ("consent_model", models.CharField(editable=False, max_length=50, null=True)),
                (
                    "consent_version",
                    models.CharField(editable=False, max_length=10, null=True),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
                (
                    "subject_visit",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_visit_tracking.subjectvisit",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Missed Visit Report",
                "verbose_name_plural": "historical Missed Visit Report",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalSubjectVisit",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                ("visit_code", models.CharField(editable=False, max_length=25, null=True)),
                (
                    "visit_code_sequence",
                    models.IntegerField(
                        blank=True,
                        default=0,
                        help_text="An integer to represent the sequence of additional appointments relative to the base appointment, 0, needed to complete data collection for the timepoint. (NNNN.0)",
                        null=True,
                        verbose_name="Sequence",
                    ),
                ),
                (
                    "visit_schedule_name",
                    models.CharField(
                        editable=False,
                        help_text='the name of the visit schedule used to find the "schedule"',
                        max_length=25,
                    ),
                ),
                ("schedule_name", models.CharField(editable=False, max_length=25)),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=edc_utils.date.get_utcnow,
                        help_text="Date and time of this report",
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.validators.date.datetime_not_future,
                        ],
                        verbose_name="Report date and time",
                    ),
                ),
                (
                    "reason_unscheduled",
                    models.CharField(
                        choices=[
                            ("patient_unwell_outpatient", "Patient unwell (outpatient)"),
                            ("patient_hospitalised", "Patient hospitalised"),
                            ("OTHER", "Other"),
                            ("N/A", "Not applicable"),
                        ],
                        default="N/A",
                        max_length=25,
                        verbose_name="If 'unscheduled', provide the reason for the unscheduled visit",
                    ),
                ),
                (
                    "reason_unscheduled_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=25,
                        null=True,
                        verbose_name='If the reason for the unscheduled visit is "other", specify',
                    ),
                ),
                (
                    "reason_missed_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=25,
                        null=True,
                        verbose_name='If the reason for the missed visit is "other", specify',
                    ),
                ),
                (
                    "study_status",
                    models.CharField(
                        max_length=50,
                        null=True,
                        verbose_name="What is the participant's current study status",
                    ),
                ),
                (
                    "require_crfs",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No")],
                        default="Yes",
                        max_length=10,
                        verbose_name="Are scheduled data being submitted with this visit?",
                    ),
                ),
                (
                    "info_source_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name='If "Other" source of information, specify',
                    ),
                ),
                (
                    "survival_status",
                    models.CharField(
                        choices=[
                            ("alive", "Alive"),
                            ("dead", "Deceased"),
                            ("unknown", "Unknown"),
                            ("N/A", "Not applicable (if missed)"),
                        ],
                        default="alive",
                        help_text="If YES, submit Death report",
                        max_length=10,
                        null=True,
                        verbose_name="Participant's survival status",
                    ),
                ),
                (
                    "last_alive_date",
                    models.DateField(
                        blank=True,
                        help_text="Will be validated on the Death report",
                        null=True,
                        validators=[
                            edc_protocol.validators.date_not_before_study_start,
                            edc_model.validators.date.date_not_future,
                        ],
                        verbose_name="Date participant last known alive",
                    ),
                ),
                (
                    "comments",
                    models.TextField(
                        blank=True,
                        max_length=250,
                        null=True,
                        verbose_name="Comment if any additional pertinent information about the participant",
                    ),
                ),
                (
                    "document_status",
                    models.CharField(
                        choices=[
                            ("INCOMPLETE", "Incomplete (some data pending)"),
                            ("COMPLETE", "Complete"),
                        ],
                        default="COMPLETE",
                        help_text="If some data is still pending, flag as incomplete",
                        max_length=25,
                        verbose_name="Document status",
                    ),
                ),
                (
                    "document_status_comments",
                    models.TextField(
                        blank=True,
                        help_text="for example, why some data is still pending",
                        null=True,
                        verbose_name="Any comments related to status of this document",
                    ),
                ),
                ("consent_model", models.CharField(editable=False, max_length=50, null=True)),
                (
                    "consent_version",
                    models.CharField(editable=False, max_length=10, null=True),
                ),
                (
                    "reason",
                    models.CharField(
                        choices=[
                            ("scheduled", "Scheduled visit/contact"),
                            ("unscheduled", "Unscheduled visit/contact"),
                            ("missed", "Missed visit"),
                            (
                                "lost",
                                "Lost to follow-up (use only when taking subject off study)",
                            ),
                            ("deferred", "Deferred"),
                            ("completed protocol", "Completed protocol"),
                        ],
                        max_length=25,
                    ),
                ),
                (
                    "reason_missed",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("timepoint", "Missed timepoint"),
                            ("OTHER", "Other"),
                            ("N/A", "Not applicable"),
                        ],
                        max_length=35,
                        null=True,
                        verbose_name="If 'missed', provide the reason for the missed visit",
                    ),
                ),
                (
                    "info_source",
                    models.CharField(
                        choices=[
                            ("participant", "1. Clinic visit with participant"),
                            ("other_contact", "2. Other contact with participant"),
                            (
                                "other_doctor",
                                "3. Contact with external health care provider/medical doctor",
                            ),
                            (
                                "family",
                                "4. Contact with family or designated person who can provide information",
                            ),
                            ("chart", "5. Hospital chart or other medical record"),
                            ("OTHER", "9. Other"),
                        ],
                        max_length=25,
                        verbose_name="What is the main source of this information?",
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "appointment",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_appointment.appointment",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical subject visit",
                "verbose_name_plural": "historical subject visits",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="SubjectVisitMissed",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, default=django_audit_fields.models.audit_model_mixin.utcnow
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=edc_utils.date.get_utcnow,
                        help_text="If reporting today, use today's date/time, otherwise use the date/time this information was reported.",
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.validators.date.datetime_not_future,
                        ],
                        verbose_name="Report Date",
                    ),
                ),
                (
                    "survival_status",
                    models.CharField(
                        choices=[
                            ("alive", "Alive"),
                            ("dead", "Deceased"),
                            ("unknown", "Unknown"),
                        ],
                        help_text="If deceased, complete the death report",
                        max_length=25,
                        verbose_name="Survival status",
                    ),
                ),
                (
                    "contact_attempted",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No")],
                        help_text="Not including pre-appointment reminders",
                        max_length=25,
                        verbose_name="Were any attempts made to contact the participant since the expected appointment date?",
                    ),
                ),
                (
                    "contact_attempts_count",
                    models.IntegerField(
                        blank=True,
                        help_text="Not including pre-appointment reminders. Multiple attempts on the same day count as a single attempt.",
                        null=True,
                        validators=[django.core.validators.MinValueValidator(1)],
                        verbose_name="Number of attempts made to contact participantsince the expected appointment date",
                    ),
                ),
                (
                    "contact_attempts_explained",
                    models.TextField(
                        blank=True,
                        null=True,
                        verbose_name="If contact not made and less than 3 attempts, please explain",
                    ),
                ),
                (
                    "contact_last_date",
                    models.DateField(
                        blank=True,
                        null=True,
                        validators=[
                            edc_model.validators.date.date_not_future,
                            edc_protocol.validators.date_not_before_study_start,
                        ],
                        verbose_name="Date of last telephone contact/attempt",
                    ),
                ),
                (
                    "contact_made",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No"), ("N/A", "Not applicable")],
                        default="N/A",
                        max_length=25,
                        verbose_name="Was contact finally made with the participant?",
                    ),
                ),
                (
                    "missed_reasons_other",
                    edc_model.models.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name="If other, please specify ...",
                    ),
                ),
                (
                    "ltfu",
                    models.CharField(
                        choices=[("Yes", "Yes"), ("No", "No"), ("N/A", "Not applicable")],
                        default="No",
                        help_text="If 'Yes', complete the Loss to Follow up form",
                        max_length=15,
                        verbose_name="Has the participant met the protocol criteria for lost to follow up?",
                    ),
                ),
                (
                    "comment",
                    models.TextField(
                        blank=True,
                        null=True,
                        verbose_name="Please provide further details, if any",
                    ),
                ),
                ("consent_model", models.CharField(editable=False, max_length=50, null=True)),
                (
                    "consent_version",
                    models.CharField(editable=False, max_length=10, null=True),
                ),
                (
                    "missed_reasons",
                    models.ManyToManyField(
                        blank=True,
                        related_name="default_missed_reasons",
                        to="edc_visit_tracking.subjectvisitmissedreasons",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
                (
                    "subject_visit",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="edc_subject_visit",
                        to="edc_visit_tracking.subjectvisit",
                    ),
                ),
            ],
            options={
                "verbose_name": "Missed Visit Report",
                "verbose_name_plural": "Missed Visit Report",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["subject_visit", "site", "id"],
                        name="edc_visit_t_subject_8fc4e6_idx",
                    ),
                    models.Index(
                        fields=["subject_visit", "report_datetime"],
                        name="edc_visit_t_subject_36f1fc_idx",
                    ),
                ],
            },
            managers=[
                ("objects", edc_visit_tracking.managers.CrfModelManager()),
                ("on_site", edc_visit_tracking.managers.CrfCurrentSiteManager()),
            ],
        ),
        migrations.AddIndex(
            model_name="subjectvisit",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "report_datetime",
                ],
                name="edc_visit_t_subject_413eb1_idx",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="subjectvisit",
            unique_together={
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                ),
                (
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "report_datetime",
                ),
            },
        ),
    ]



from unittest.mock import patch

from dateutil.relativedelta import relativedelta
from django import forms
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import ALIVE, OTHER, YES
from edc_facility.import_holidays import import_holidays
from edc_form_validators import APPLICABLE_ERROR, REQUIRED_ERROR
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from edc_visit_tracking.form_validators import VisitFormValidator
from edc_visit_tracking.models import SubjectVisit

from ..helper import Helper
from ..visit_schedule import visit_schedule1, visit_schedule2


class TestSubjectVisitFormValidator(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        self.helper.consent_and_put_on_schedule()
        self.appointment = Appointment.objects.all().order_by("timepoint_datetime")[0]

    def test_form_validator_ok(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        cleaned_data = dict(
            appointment=appointment,
            reason=SCHEDULED,
            is_present=YES,
            survival_status=ALIVE,
            last_alive_date=get_utcnow().date(),
        )
        form_validator = VisitFormValidator(cleaned_data=cleaned_data, instance=subject_visit)
        form_validator.validate()

    def test_visit_code_reason_with_visit_code_sequence_0(self):
        cleaned_data = {"appointment": self.appointment, "reason": UNSCHEDULED}
        form_validator = VisitFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("reason", form_validator._errors)

    @patch("edc_appointment.form_validators.utils.get_appointment_url")
    def test_visit_code_reason_with_visit_code_sequence_1(self, mock_urlnames):
        mock_urlnames.return_value = ""
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)

        opts = self.appointment.__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts.update(
            visit_code_sequence=1,
            appt_datetime=self.appointment.appt_datetime + relativedelta(days=1),
        )
        appointment = Appointment.objects.create(**opts)

        cleaned_data = {"appointment": appointment, "reason": SCHEDULED}
        form_validator = VisitFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("reason", form_validator._errors)

    def test_visit_code_reason_with_visit_code_sequence_2(self):
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)

        opts = self.appointment.__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts.update(
            visit_code_sequence=1,
            appt_datetime=self.appointment.appt_datetime + relativedelta(days=1),
        )
        Appointment.objects.create(**opts)
        opts.update(
            visit_code_sequence=2,
            appt_datetime=self.appointment.appt_datetime + relativedelta(days=2),
        )
        appointment = Appointment.objects.create(**opts)

        cleaned_data = {"appointment": appointment, "reason": SCHEDULED}
        form_validator = VisitFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn(
            "Previous visit report required",
            ",".join([str(e) for e in form_validator._errors.values()]),
        )
        self.assertIn("1000.1", ",".join([str(e) for e in form_validator._errors.values()]))

    def test_reason_missed(self):
        options = {
            "appointment": self.appointment,
            "reason": MISSED_VISIT,
            "reason_missed": None,
        }
        form_validator = VisitFormValidator(cleaned_data=options)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("reason_missed", form_validator._errors)

    @patch("edc_appointment.form_validators.utils.get_appointment_url")
    def test_reason_unscheduled(self, mock_urlnames):
        mock_urlnames.return_value = ""
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)

        # create unscheduled appointment (XXXX.1)
        opts = self.appointment.__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts.update(
            appt_datetime=self.appointment.appt_datetime + relativedelta(days=1),
            visit_code_sequence=1,
        )
        appointment = Appointment.objects.create(**opts)
        # start subject visit form with unscheduled appointment
        options = {
            "appointment": appointment,
            "reason": UNSCHEDULED,
            "reason_unscheduled": None,
        }
        form_validator = VisitFormValidator(cleaned_data=options)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("reason_unscheduled", form_validator._errors)
        self.assertIn(APPLICABLE_ERROR, form_validator._error_codes)

    @patch("edc_appointment.form_validators.utils.get_appointment_url")
    def test_reason_unscheduled_other(self, mock_urlnames):
        mock_urlnames.return_value = ""
        SubjectVisit.objects.create(appointment=self.appointment, reason=SCHEDULED)

        opts = self.appointment.__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts.update(
            visit_code_sequence=1,
            appt_datetime=self.appointment.appt_datetime + relativedelta(days=1),
        )
        appointment = Appointment.objects.create(**opts)

        options = {
            "appointment": appointment,
            "reason": UNSCHEDULED,
            "reason_unscheduled": OTHER,
            "reason_unscheduled_other": None,
        }
        form_validator = VisitFormValidator(cleaned_data=options)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("reason_unscheduled_other", form_validator._errors)
        self.assertIn(REQUIRED_ERROR, form_validator._error_codes)

    def test_info_source_other(self):
        options = {
            "appointment": self.appointment,
            "info_source": OTHER,
            "info_source_other": None,
        }
        form_validator = VisitFormValidator(cleaned_data=options)
        try:
            form_validator.validate()
        except forms.ValidationError:
            pass
        self.assertIn("info_source_other", form_validator._errors)
        self.assertIn(REQUIRED_ERROR, form_validator._error_codes)
